==> sys-20.5.19.jar: dev/polymeris/sys/db.clj <==
(ns dev.polymeris.sys.db
  (:require
    [integrant.core :as ig]
    [clojure.tools.logging :as log]
    [clojure.spec.alpha :as s]
    [ragtime.jdbc]
    [ragtime.repl]
    [next.jdbc :as jdbc]
    [jsonista.core :as json])
  (:import
    (java.sql Array Date Time Timestamp PreparedStatement)
    (org.postgresql.util PGobject)
    (clojure.lang IPersistentMap IPersistentSet ISeq)))


(s/def ::uri string?)
(s/def ::migrations-dir string?)
(s/def ::pool-size pos-int?)

(s/def ::connection #(not (.isClosed %)))

(s/def ::database
  (s/keys :req-un [::uri ::migrations-dir ::connection]))

(defn migration-config
  [{:keys [uri migrations-dir]}]
  {:datastore (ragtime.jdbc/sql-database {:connection-uri uri})
   :migrations (ragtime.jdbc/load-resources migrations-dir)})

(defn migrate!
  [database]
  (-> (migration-config database)
      ragtime.repl/migrate))

(defmethod ig/pre-init-spec ::database [_]
  (s/keys :req-un [::uri]
          :opt-un [::migrations-dir ::pool-size]))

(defmethod ig/init-key ::database
  [_ {:keys [uri migrations-dir pool-size]
      :or {migrations-dir "migrations"
           pool-size 5}}]
  (let [connection (jdbc/get-connection {:jdbc-url uri
                                         :maximum-pool-size pool-size})]
    (log/info "Database connection established")
    {:uri uri
     :migrations-dir migrations-dir
     :connection connection}))

(defmethod ig/halt-key! ::database
  [_ {:keys [connection] :as database}]
  (when (s/valid? ::database database)
    (.close connection)
    (log/info "Database connection destroyed")))


(def ^:private json-mapper
  (json/object-mapper
    {:encode-key-fn name
     :decode-key-fn keyword}))

(defn pgobj->clj [^PGobject pgobj]
  (let [type (.getType pgobj)
        value (.getValue pgobj)]
    (case type
      "json" (json/read-value value json-mapper)
      "jsonb" (json/read-value value json-mapper)
      "citext" (str value)
      value)))

(extend-protocol next.jdbc.result-set/ReadableColumn
  Timestamp
  (read-column-by-label [^Timestamp v _]
    (.toLocalDateTime v))
  (read-column-by-index [^Timestamp v _2 _3]
    (.toLocalDateTime v))
  Date
  (read-column-by-label [^Date v _]
    (.toLocalDate v))
  (read-column-by-index [^Date v _2 _3]
    (.toLocalDate v))
  Time
  (read-column-by-label [^Time v _]
    (.toLocalTime v))
  (read-column-by-index [^Time v _2 _3]
    (.toLocalTime v))
  Array
  (read-column-by-label [^Array v _]
    (vec (.getArray v)))
  (read-column-by-index [^Array v _2 _3]
    (vec (.getArray v)))
  PGobject
  (read-column-by-label [^PGobject pgobj _]
    (pgobj->clj pgobj))
  (read-column-by-index [^PGobject pgobj _2 _3]
    (pgobj->clj pgobj)))

(defn clj->jsonb-pgobj [value]
  (doto (PGobject.)
    (.setType "jsonb")
    (.setValue (json/write-value-as-string value json-mapper))))

(extend-protocol next.jdbc.prepare/SettableParameter
  IPersistentMap
  (set-parameter [^IPersistentMap v ^PreparedStatement stmt ^long idx]
    (.setObject stmt idx (clj->jsonb-pgobj v)))
  IPersistentSet
  (set-parameter [^IPersistentMap v ^PreparedStatement stmt ^long idx]
    (.setObject stmt idx (clj->jsonb-pgobj v)))
  ISeq
  (set-parameter [^IPersistentMap v ^PreparedStatement stmt ^long idx]
    (.setObject stmt idx (clj->jsonb-pgobj v))))

==> sys-20.5.21.jar: dev/polymeris/sys/db.clj <==
(ns dev.polymeris.sys.db
  (:require
    [integrant.core :as ig]
    [clojure.tools.logging :as log]
    [clojure.spec.alpha :as s]
    [ragtime.jdbc]
    [ragtime.repl]
    [next.jdbc :as jdbc]
    [jsonista.core :as json])
  (:import
    (java.sql Array Date Time Timestamp PreparedStatement)
    (org.postgresql.util PGobject)
    (clojure.lang IPersistentMap IPersistentSet ISeq)))


(s/def ::uri string?)
(s/def ::migrations-dir string?)

(s/def ::connection #(not (.isClosed %)))

(s/def ::database
  (s/keys :req-un [::uri ::migrations-dir ::connection]))

(defn migration-config
  [{:keys [uri migrations-dir]}]
  {:datastore (ragtime.jdbc/sql-database {:connection-uri uri})
   :migrations (ragtime.jdbc/load-resources migrations-dir)})

(defn migrate!
  [database]
  (-> (migration-config database)
      ragtime.repl/migrate))

(defmethod ig/pre-init-spec ::database [_]
  (s/keys :req-un [::uri]
          :opt-un [::migrations-dir]))

(defmethod ig/init-key ::database
  [_ {:keys [uri migrations-dir]
      :or {migrations-dir "migrations"}}]
  (let [connection (jdbc/get-connection uri)]
    (log/info "Database connection established")
    {:uri uri
     :migrations-dir migrations-dir
     :connection connection}))

(defmethod ig/halt-key! ::database
  [_ {:keys [connection] :as database}]
  (when (s/valid? ::database database)
    (.close connection)
    (log/info "Database connection destroyed")))


(def ^:private json-mapper
  (json/object-mapper
    {:encode-key-fn name
     :decode-key-fn keyword}))

(defn pgobj->clj [^PGobject pgobj]
  (let [type (.getType pgobj)
        value (.getValue pgobj)]
    (case type
      "json" (json/read-value value json-mapper)
      "jsonb" (json/read-value value json-mapper)
      "citext" (str value)
      value)))

(extend-protocol next.jdbc.result-set/ReadableColumn
  Timestamp
  (read-column-by-label [^Timestamp v _]
    (.toLocalDateTime v))
  (read-column-by-index [^Timestamp v _2 _3]
    (.toLocalDateTime v))
  Date
  (read-column-by-label [^Date v _]
    (.toLocalDate v))
  (read-column-by-index [^Date v _2 _3]
    (.toLocalDate v))
  Time
  (read-column-by-label [^Time v _]
    (.toLocalTime v))
  (read-column-by-index [^Time v _2 _3]
    (.toLocalTime v))
  Array
  (read-column-by-label [^Array v _]
    (vec (.getArray v)))
  (read-column-by-index [^Array v _2 _3]
    (vec (.getArray v)))
  PGobject
  (read-column-by-label [^PGobject pgobj _]
    (pgobj->clj pgobj))
  (read-column-by-index [^PGobject pgobj _2 _3]
    (pgobj->clj pgobj)))

(defn clj->jsonb-pgobj [value]
  (doto (PGobject.)
    (.setType "jsonb")
    (.setValue (json/write-value-as-string value json-mapper))))

(extend-protocol next.jdbc.prepare/SettableParameter
  IPersistentMap
  (set-parameter [^IPersistentMap v ^PreparedStatement stmt ^long idx]
    (.setObject stmt idx (clj->jsonb-pgobj v)))
  IPersistentSet
  (set-parameter [^IPersistentMap v ^PreparedStatement stmt ^long idx]
    (.setObject stmt idx (clj->jsonb-pgobj v)))
  ISeq
  (set-parameter [^IPersistentMap v ^PreparedStatement stmt ^long idx]
    (.setObject stmt idx (clj->jsonb-pgobj v))))
